@@ -1,18 +1,15 @@
-from __future__ import absolute_import
-from __future__ import unicode_literals
-
+import _thread as thread
 import logging
 import operator
 import sys
+from queue import Empty
+from queue import Queue
 from threading import Lock
 from threading import Semaphore
 from threading import Thread
 
 from docker.errors import APIError
 from docker.errors import ImageNotFound
-from six.moves import _thread as thread
-from six.moves.queue import Empty
-from six.moves.queue import Queue
 
 from compose.cli.colors import green
 from compose.cli.colors import red
@@ -21,7 +18,6 @@ from compose.const import PARALLEL_LIMIT
 from compose.errors import HealthCheckFailed
 from compose.errors import NoHealthCheckConfigured
 from compose.errors import OperationFailedError
-from compose.utils import get_output_stream
 
 
 log = logging.getLogger(__name__)
@@ -29,7 +25,7 @@ log = logging.getLogger(__name__)
 STOP = object()
 
 
-class GlobalLimit(object):
+class GlobalLimit:
     """Simple class to hold a global semaphore limiter for a project. This class
     should be treated as a singleton that is instantiated when the project is.
     """
@@ -43,14 +39,17 @@ class GlobalLimit(object):
         cls.global_limiter = Semaphore(value)
 
 
-def parallel_execute_watch(events, writer, errors, results, msg, get_name):
+def parallel_execute_watch(events, writer, errors, results, msg, get_name, fail_check):
     """ Watch events from a parallel execution, update status and fill errors and results.
         Returns exception to re-raise.
     """
     error_to_reraise = None
     for obj, result, exception in events:
         if exception is None:
-            writer.write(msg, get_name(obj), 'done', green)
+            if fail_check is not None and fail_check(obj):
+                writer.write(msg, get_name(obj), 'failed', red)
+            else:
+                writer.write(msg, get_name(obj), 'done', green)
             results.append(result)
         elif isinstance(exception, ImageNotFound):
             # This is to bubble up ImageNotFound exceptions to the client so we
@@ -72,15 +71,17 @@ def parallel_execute_watch(events, writer, errors, results, msg, get_name):
     return error_to_reraise
 
 
-def parallel_execute(objects, func, get_name, msg, get_deps=None, limit=None):
+def parallel_execute(objects, func, get_name, msg, get_deps=None, limit=None, fail_check=None):
     """Runs func on objects in parallel while ensuring that func is
     ran on object only after it is ran on all its dependencies.
 
     get_deps called on object must return a collection with its dependencies.
     get_name called on object must return its name.
+    fail_check is an additional failure check for cases that should display as a failure
+        in the CLI logs, but don't raise an exception (such as attempting to start 0 containers)
     """
     objects = list(objects)
-    stream = get_output_stream(sys.stderr)
+    stream = sys.stderr
 
     if ParallelStreamWriter.instance:
         writer = ParallelStreamWriter.instance
@@ -96,7 +97,9 @@ def parallel_execute(objects, func, get_name, msg, get_deps=None, limit=None):
 
     errors = {}
     results = []
-    error_to_reraise = parallel_execute_watch(events, writer, errors, results, msg, get_name)
+    error_to_reraise = parallel_execute_watch(
+        events, writer, errors, results, msg, get_name, fail_check
+    )
 
     for obj_name, error in errors.items():
         stream.write("\nERROR: for {}  {}\n".format(obj_name, error))
@@ -111,7 +114,7 @@ def _no_deps(x):
     return []
 
 
-class State(object):
+class State:
     """
     Holds the state of a partially-complete parallel operation.
 
@@ -133,7 +136,7 @@ class State(object):
         return set(self.objects) - self.started - self.finished - self.failed
 
 
-class NoLimit(object):
+class NoLimit:
     def __enter__(self):
         pass
 
@@ -249,7 +252,7 @@ class UpstreamError(Exception):
     pass
 
 
-class ParallelStreamWriter(object):
+class ParallelStreamWriter:
     """Write out messages for operations happening in parallel.
 
     Each operation has its own line, and ANSI code characters are used
